#include "zftp.h"
#include <stdlib.h>
#include <stdio.h>
#ifdef WIN32
#  include <direct.h>
#endif

bool ParseArg(ZFTP *pzftp, int argc, char **argv);

int main(int argc, char **argv)
{
	ZFTP zftp;
	Init(&zftp);
	
	if ( !ParseArg(&zftp, argc, argv) )
		return 0;

	return Process(&zftp);
}

bool ParseSubArg(const char *pIn, char *rpszUserID, char *rpszServer, char *rpszPath)
{
	const char *p = strchr(pIn, '@');
	
	if ( !p )
		return false;

	strncpy(rpszUserID, pIn, p-pIn);

	pIn = p+1;

	if ( (p=strchr(pIn, ':')) == NULL )
		return false;

	strncpy(rpszServer, pIn, p-pIn);

	pIn = p;
	
	sprintf(rpszPath, "%s", p+1);
	return true;
}

bool ParseArg(ZFTP *pzftp, int argc, char **argv)
{
	int i = 0, n = argc;
	char szLocalPath[MAX_PATH];
	char szRemotePath[MAX_PATH];

	InitList(&pzftp->m_lstPath);
	memset(szLocalPath, 0, MAX_PATH);
	memset(szRemotePath, 0, MAX_PATH);

	for ( i = 1; i < n; i++ )
	{
		if ( memicmp(argv[i], "-pass:", 6) == 0 )
		{
			sprintf(pzftp->m_szUserPW, "%s", argv[i]+6);
		}
		else if ( memicmp(argv[i], "-?", 2) == 0 )
		{
			PrintHelp(true);
		}
		else if ( memicmp(argv[i], "-v", 2) == 0 )
		{
			PrintVer(true);
		}
		else if ( memicmp(argv[i], "-put", 4) == 0 )
		{
			pzftp->m_cmd = PUT;
			sprintf(szLocalPath, "%s", argv[i+1]);

			if ( szLocalPath[0] == 0 )
				PrintHelp(true);
								
			if ( strchr(argv[i+2], ':') == NULL )
			{
				// on case zftp -put c:\test.zip userid@host.com

				const char *pIn = argv[i+2];
				const char *p = strchr(pIn, '@');
				
				if ( !p )
					PrintHelp(true);

				strncpy(pzftp->m_szUserID, pIn, p-pIn);
				pIn = p+1;
				sprintf(pzftp->m_szServer, "%s", pIn);

				
				if ( memicmp(szLocalPath, "*.", 2) == 0 )
				{
					char szPath[MAX_PATH];
					getcwd(szPath, MAX_PATH);
					InitCursor(pzftp, szPath, szLocalPath);
					
					while ( GetNextPath(pzftp, szPath) )
					{
						char szFileName[MAX_PATH];
						PATH *path = (PATH*)malloc(sizeof(PATH));
						sprintf(path->m_szLocalPath, "%s", szPath);
						ExtractFileName(path->m_szLocalPath, szFileName);
						sprintf(path->m_szRemotePath, "%s", szFileName);
						PushBack(&pzftp->m_lstPath, (void *)path);
					}
				}
				else
				{
					char szFileName[MAX_PATH];
					PATH *path = (PATH*)malloc(sizeof(PATH));
					sprintf(path->m_szLocalPath, "%s", szLocalPath);

					if ( access(path->m_szLocalPath, 0) != 0 )
					{
						WriteLog(pzftp, "[main.c %d] %s, %s\n", __LINE__, strerror(errno), path->m_szLocalPath);
						exit(1);
					}

					if ( HasDDIR(szLocalPath) )
					{
						ExtractFileName(szLocalPath, szFileName);
						sprintf(path->m_szRemotePath, "%s", szFileName);
					}
					else
					{
						sprintf(path->m_szRemotePath, "%s", szLocalPath);
					}

					PushBack(&pzftp->m_lstPath, (void*)path);
				}
			}
			else
			{
				PATH *path = (PATH*)malloc(sizeof(PATH));
				if ( !ParseSubArg(argv[i+2], pzftp->m_szUserID, pzftp->m_szServer, szRemotePath) )
					PrintHelp(true);

				sprintf(path->m_szLocalPath, "%s", szLocalPath);
				sprintf(path->m_szRemotePath, "%s", szRemotePath);
				PushBack(&pzftp->m_lstPath, (void*)path);
			}

			break;
		}
		else if ( memicmp(argv[i], "-get", 4) == 0 )
		{
			bool bDir = false;
			PATH *path = (PATH*)malloc(sizeof(PATH));

			memset(path->m_szLocalPath, 0, MAX_PATH);
			memset(path->m_szRemotePath, 0, MAX_PATH);

			pzftp->m_cmd = GET;

			if ( i+2 < argc )
			{
				sprintf(szLocalPath, "%s", argv[i+2]);
				if ( strchr(szLocalPath, DDIR) == NULL )
				{
					// on case -get aswip@myftp.hanafos.com:url.txt url.txt
					char szCwd[MAX_PATH];
					memset(szCwd, 0, MAX_PATH);
					getcwd(szCwd, MAX_PATH);
					EnsureDDIR(szCwd, DDIR);
					sprintf(szLocalPath, "%s%s", szCwd, argv[i+2]);
				}

				if ( access(szLocalPath, 0) == 0 && IsDir(szLocalPath) )
					bDir = true;
			}
			else
			{
				// on case -get aswip@myftp.hanafos.com:url.txt
			}
			
			if ( !ParseSubArg(argv[i+1], pzftp->m_szUserID, pzftp->m_szServer, szRemotePath) )
				PrintHelp(true);

			if ( bDir )
			{
				char szFileName[MAX_PATH];
				memset(szFileName, 0, MAX_PATH);
				ExtractFileName(szRemotePath, szFileName);
				strcat(szLocalPath, szFileName);
			}
			else if ( szLocalPath[0] == 0 )
			{
				char szFileName[MAX_PATH];
				char szCwd[MAX_PATH];

				memset(szFileName, 0, MAX_PATH);
				memset(szCwd, 0, MAX_PATH);
				
				if ( HasDDIR(szRemotePath) )
				{
					ExtractFileName(szRemotePath, szFileName);
					strcat(szLocalPath, szFileName);
				}
				else
				{
					getcwd(szCwd, MAX_PATH);
					EnsureDDIR(szCwd, DDIR);
					sprintf(szLocalPath, "%s%s", szCwd, szRemotePath);
				}
			}

			sprintf(path->m_szLocalPath, "%s", szLocalPath);
			sprintf(path->m_szRemotePath, "%s", szRemotePath);
			PushBack(&pzftp->m_lstPath, (void*)path);
			break;
		}
	}

	if ( pzftp->m_cmd != PUT && pzftp->m_cmd != GET )
		PrintHelp(true);
	
	if ( pzftp->m_lstPath.m_nCount == 0 )
	{
		if ( szLocalPath[0] == 0 || szRemotePath[0] == 0 )
			PrintHelp(true);
	}

	return true;
}

