/**
	@file		zftp.c
	@author		aswip78@gmail.com
	@date		2008 / 02 / 06
*/
#include "zftp.h"
#include <fcntl.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>

#ifdef WIN32
#  include <direct.h>
#  include <winsock2.h>
#else
#  include <string.h>
#endif

#ifdef WIN32
typedef int socklen_t;
#endif

#ifndef WIN32
int closesocket(int nSockFD)
{
	return close(nSockFD);
}
#endif

#ifdef WIN32
void sleep(int nSec)
{
	Sleep(nSec*1000);
}
#endif

void Init(ZFTP *pzftp)
{
	pzftp->m_pszSendBuffer = (char *) malloc (MAX_SOCK_BUFFER+10);
	pzftp->m_pszRecvBuffer = (char *) malloc (MAX_SOCK_BUFFER+10);
	pzftp->m_pszLineBuffer = (char *) malloc (MAX_SOCK_BUFFER+10);
	pzftp->m_szError[0] = 0;
	pzftp->m_nTimeout = 3600;
	pzftp->m_nPort = 21;
	pzftp->m_sysType = UNIX;

#ifdef WIN32
	pzftp->m_hDir = NULL;
	pzftp->m_bFirst = true;
#else
	pzftp->m_pDir = NULL;
#endif

	pzftp->m_nBufIdx = 0;
	pzftp->m_nBufLeng = 0;
	pzftp->m_bSilent = false;

	InitList(&pzftp->m_lstLocalIP);
	memset(pzftp->m_szUserID, 0x00, sizeof(pzftp->m_szUserID)/sizeof(char));
	memset(pzftp->m_szUserPW, 0x00, sizeof(pzftp->m_szUserPW)/sizeof(char));
	memset(pzftp->m_szServer, 0x00, sizeof(pzftp->m_szServer)/sizeof(char));
	memset(pzftp->m_szDir, 0x00, sizeof(pzftp->m_szDir)/sizeof(char));
	memset(pzftp->m_szLocalIP, 0x00, sizeof(pzftp->m_szLocalIP)/sizeof(char));

#ifdef WIN32
	InitializeCriticalSection(&pzftp->m_Lock);
#else
	pthread_mutex_init(&pzftp->m_Lock, NULL);
#endif	
}

void UnInit(ZFTP *pzftp)
{
	free(pzftp->m_pszSendBuffer);
	free(pzftp->m_pszRecvBuffer);
	free(pzftp->m_pszLineBuffer);

#ifdef WIN32
	DeleteCriticalSection(&pzftp->m_Lock);
#else
	pthread_mutex_destroy(&pzftp->m_Lock);
#endif

	FreeList(&pzftp->m_lstLocalIP);
	FreeList(&pzftp->m_lstPath);
}

void Lock(ZFTP *pzftp)
{
#ifdef WIN32
	EnterCriticalSection(&pzftp->m_Lock);
#else
	pthread_mutex_lock(&pzftp->m_Lock);
#endif
}

void UnLock(ZFTP *pzftp)
{
#ifdef WIN32
	LeaveCriticalSection(&pzftp->m_Lock);
#else
	pthread_mutex_unlock(&pzftp->m_Lock);
#endif
}

const char *GetVerStr()
{
	static const char *ver = "v1.00.02";
	return ver;
}

void SetTimeout(ZFTP *pzftp, int nTimeout)
{
	pzftp->m_nTimeout = nTimeout;
}

void SetServer(ZFTP *pzftp, const char *pszServer, const int nPort)
{
	sprintf(pzftp->m_szServer, "%s", pszServer);
	pzftp->m_nPort = (unsigned short)nPort;
}

void SetAccount(ZFTP *pzftp, const char *pszUserID, const char *pszUserPW)
{
	sprintf(pzftp->m_szUserID, "%s", pszUserID);
	sprintf(pzftp->m_szUserPW, "%s", pszUserPW);
}

int PrintHelp(bool bDirectExit)
{
#ifdef WIN32
	static const char *pszApp = "zftp.exe";
#else
	static const char *pszApp = "zftp";
#endif

	printf("%s %s\n", pszApp, GetVerStr());
	printf("\n");
	printf("zftp copies files between hosts on a network.\n"); 
	printf("It uses ftp for data transfer, and uses the same authentication.\n");
	printf("Any file name may contain a host and user specification to \n");
	printf("indicate that the file is to be copied to/from that host.\n");
	printf("Copies between two remote hosts are permitted.\n");
	printf("\n");
	printf("The options are as follows:\n");
	printf("\n");
	printf("  -put\n");
	printf("\tUpload local files to remote. and it needs two argument.\n");
 	printf("\n");
	printf("\t-put <local file path> <userid@host:remote file path>\n");
	printf("\n");
	printf("\tor\n");
	printf("\n");
	printf("\t-put <local file path> <userid@host>\n");
	printf("\n");
	printf("\tsimple usage are follows.\n");
	printf("\n");
	printf("\tzftp -put data.txt userid@host.com\n");
	printf("\tzftp -put data.txt userid@host.com:data.txt\n");
	printf("\tzftp -put c:\\data.txt userid@host.com:/home/userid/data.txt\n");
	printf("\tzftp -put *.txt userid@host.com\n");
	printf("\n");
	printf("  -get\n");
	printf("\n");
	printf("\tDownload remote files to local. and it needs two argument.\n");
	printf("\n");
	printf("\t-get <userid@host:remote file path> <local file path>\n");
	printf("\n");
	printf("\tor\n");
	printf("\n");
	printf("\t-get <userid@host:remote file path>\n");
	printf("\n");
	printf("\tsimple usage are follows.\n");
	printf("\n");
	printf("\tzftp -get userid@host.com:data.txt data.txt\n");
	printf("\tzftp -get userid@host.com:data.txt\n");
	printf("\n");
	printf("  -pass\n");
	printf("\n");
	printf("\tset password for authentication.\n");
	printf("\twhen you use this option at command line,\n"); 
	printf("\tyou don't have to type password.\n");
	printf("\n");
	printf("\tsimple usage are follows.\n");
	printf("\n");
	printf("\tzftp -pass:yourpw -put data.txt userid@host.com\n");
	printf("\tzftp -pass:yourpw -get userid@host.com:data.txt\n");
	printf("\tzftp -pass:yourpw -get userid@host.com:data.txt c:\\data.txt\n");
	printf("\tzftp -pass:yourpw -get userid@host.com:/test/data.txt c:\\data.txt\n");

	printf("\n");
	printf("  -?\n");
	printf("\tprint help\n");
	printf("\n");
	printf("  -v\n");
	printf("\tprint version\n");
	printf("\n");

	if ( bDirectExit )
		exit(1);

	return 0;
}

int PrintVer(bool bDirectExit)
{
#ifdef WIN32
	printf("ZFTP.exe %s\n", GetVerStr());
#else
	printf("ZFTP %s\n", GetVerStr());
#endif

	if ( bDirectExit )
		exit(1);

	return 0;
}

void SetScrEcho(bool bEnable)
{
#ifdef WIN32
	
#else
	struct termios buf;

	tcgetattr(0, &buf);

	if ( bEnable )
		buf.c_lflag |= ECHO;
	else
		buf.c_lflag &= ~(ECHO);

	tcsetattr(0, TCSAFLUSH, &buf);
#endif
}

int GetCH()
{
#ifdef WIN32
	return (int)getch();
#endif
	return getchar();
}

int EatPassword(char *rpszPass)
{
	char c;
	int nCnt = 0;
	
	SetScrEcho(false);
	while ( (c = (char)GetCH()) != 13 )
	{
		switch ( c )
		{
		case 10:
		case 13:
			SetScrEcho(true);
			return nCnt;
		case 8:	// BACKSPACE
			if ( nCnt > 0 )
			{
				rpszPass[nCnt-1] = 0;
				nCnt--;
			}
			break;
		default:	
			rpszPass[nCnt++] = c;
			break;
		}
	}

	SetScrEcho(true);	
	return nCnt;
}

void ExtractFileName(const char *pszPath, char *rpszFileName)
{
	const char *p = 0;
	if ( (p=strrchr(pszPath, '/')) != NULL )
	{
		sprintf(rpszFileName, "%s", p+1);
		return;
	}

	if ( (p=strrchr(pszPath, '\\')) != NULL )
	{
		sprintf(rpszFileName, "%s", p+1);
		return;
	}
}

bool IsDir(const char *pszPath)
{
	struct stat stIn;
	if ( stat(pszPath, &stIn) != 0 )
		return false;

	if ( (stIn.st_mode & S_IFMT) == S_IFDIR )
		return true;

	return false;
}

void SafeSRand(ZFTP *pzftp)
{
	static time_t tSeed = 0;
	
	if ( tSeed <= 102400 )
		tSeed = time(NULL);
	
	Lock(pzftp);
	tSeed--;
	srand(tSeed);
	UnLock(pzftp);
}

const char *GetLastErr(ZFTP *pzftp)
{
	return pzftp->m_szError;
}

int Process(ZFTP *pzftp)
{
	int i = 0, n = 0;
	int64 nTotRecvSize = 0;
	PTRNODE *ptrNext = NULL;
	PTRNODE *ptrNode = NULL;

	while ( pzftp->m_szUserPW[0] == 0 )
	{
		printf("enter %s@%s's password : ", pzftp->m_szUserID, pzftp->m_szServer);
		fflush(stdout);

		EatPassword(pzftp->m_szUserPW);
		if ( pzftp->m_szUserPW[0] != 0 )
			putchar('\n');
	}

	fflush(stdout);
	n = strlen(pzftp->m_szUserPW);

	for ( i = 0; i < n; i++ )
	{
		if ( pzftp->m_szUserPW[i] == '\r' || pzftp->m_szUserPW[i] == '\n' )
		{
			pzftp->m_szUserPW[i] = 0;
			break;
		}
	}

	if ( !LOGIN(pzftp) )
	{
		printf("fail to login, because %s\n", pzftp->m_szError);
		exit(1);
	}
	
	switch ( pzftp->m_cmd )
	{
	case GET:

		ptrNext = NULL;
		ptrNode = pzftp->m_lstPath.m_pHead;
		
		while ( ptrNode )
		{
			PATH *path = (PATH*)ptrNode->m_pData;
			SafeSRand(pzftp);
			if ( !Download(pzftp, path->m_szRemotePath, path->m_szLocalPath, &nTotRecvSize, true) )
				printf("%s\n", pzftp->m_szError);

			ptrNext = ptrNode->m_pNext;
			ptrNode = ptrNext;
		}
		break;
	case PUT:
		ptrNext = NULL;
		ptrNode = pzftp->m_lstPath.m_pHead;
		
		while ( ptrNode )
		{
			PATH *path = (PATH*)ptrNode->m_pData;
			SafeSRand(pzftp);
			if ( !Upload(pzftp, path->m_szLocalPath, path->m_szRemotePath, true) )
				printf("%s\n", pzftp->m_szError);

			ptrNext = ptrNode->m_pNext;
			ptrNode = ptrNext;
		}
		break;
	default:
		PrintHelp(true);
		break;
	}
	
	UnInit(pzftp);	
	return 0;
}


bool EnableNonBlockSock(ZFTP *pzftp, int nFD, bool bEnable)
{
#ifdef WIN32
	
	unsigned long ulParam = (bEnable == true) ? 1 : 0;
	
	if ( ioctlsocket(nFD, FIONBIO, &ulParam) != 0 )
	{
		sprintf(pzftp->m_szError, "fail to set EnableBlockSock, %s", GetSysErr(STDSOCKERR));
		return false;
	}

#else

	int nFlags = 0;

	// If the fcntl() function fails, a value of -1 is returned and errno is set to indicate the error.
	nFlags = fcntl(nFD, F_GETFL, 0);

	if ( nFlags < 0 )
	{
		sprintf(pzftp->m_szError, "fail to set EnableBlockSock, %s", strerror(errno));
		return false;
	}

	if ( bEnable )
	{
		if ( (nFlags & O_NONBLOCK) == O_NONBLOCK )
			return true;

		if ( fcntl(nFD, F_SETFL, nFlags | O_NONBLOCK) < 0 )
		{
			sprintf(pzftp->m_szError, "fail to set EnableBlockSock, %s", strerror(errno));
			return false;
		}

		return true;
	}

	if ( (nFlags & O_NONBLOCK) == O_NONBLOCK )
	{
		if ( fcntl(nFD, F_SETFL, nFlags & ~O_NONBLOCK) < 0 )
		{
			sprintf(pzftp->m_szError, "fail to set EnableBlockSock, %s", strerror(errno));
			return false;
		}
	}

#endif	
	return true;
}


bool Connect(ZFTP *pzftp)
{
	int i = 0;
	bool bRes= false;
	struct hostent *pHostEnt;

	memset(&pzftp->m_SockAddr.sin_addr, 0, sizeof(pzftp->m_SockAddr.sin_addr));
	pzftp->m_SockAddr.sin_family = AF_INET;
	pzftp->m_SockAddr.sin_port = htons(pzftp->m_nPort);

	if ( (pHostEnt = gethostbyname(pzftp->m_szServer)) == NULL )
	{
		pzftp->m_SockAddr.sin_addr.s_addr = inet_addr(pzftp->m_szServer);
		if ( pzftp->m_SockAddr.sin_addr.s_addr == 0xffffffff )
		{
			sprintf(pzftp->m_szError, "In Connect(), Invalid IP Format");
			return false;
		}
		
		return ConnectAddr(pzftp, pzftp->m_nFD, pzftp->m_szServer, pzftp->m_nPort, &pzftp->m_SockAddr);
	}

	for ( i=0; pHostEnt->h_addr_list[i]; i++ )
	{
		memset(&(pzftp->m_SockAddr.sin_addr), 0, sizeof((pzftp->m_SockAddr.sin_addr)));
		memcpy(&(pzftp->m_SockAddr.sin_addr), pHostEnt->h_addr_list[i], pHostEnt->h_length);
		
		if ( (bRes = ConnectAddr(pzftp, pzftp->m_nFD, pzftp->m_szServer, pzftp->m_nPort, &pzftp->m_SockAddr)) == true )
			return true;

		closesocket(pzftp->m_nFD);
		
		if((pzftp->m_nFD = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
		{
			sprintf(pzftp->m_szError, "In Connect(), Fail to create socket, %s", GetSysErr(STDSOCKERR));
			return false;
		}
	}

	return false;
}

bool ConnectAddr(ZFTP *pzftp, int nFD, const char *pszServer, int nPort, struct sockaddr_in *psa)
{
	int n = 0;
	fd_set rset;
	fd_set wset;
	struct timeval timeout;

	if ( pzftp->m_nTimeout <= 0 )
	{
		// blocking mode 
		if ( connect(nFD, (struct sockaddr *) psa, sizeof(struct sockaddr_in)) == 0 )
			return true;

		sprintf(pzftp->m_szError, "In Connect(), Fail to connect to %s, %s", pszServer, GetSysErr(STDSOCKERR));
		return false;
	}

	timeout.tv_sec = pzftp->m_nTimeout;
	timeout.tv_usec = 0L;

	EnableNonBlockSock(pzftp, nFD, true);

	if ( (n = connect(nFD, (struct sockaddr *) psa, sizeof(*psa))) < 0 )
	{
#ifdef WIN32
		if ( WSAGetLastError() != WSAEWOULDBLOCK )
#else
		if ( errno != EINPROGRESS )
#endif
		{
			EnableNonBlockSock(pzftp, nFD, false);
			sprintf(pzftp->m_szError, "In Connect(), fail to connect to %s : %s", pzftp->m_szServer, GetSysErr(STDSOCKERR));
			return false;
		}
	}

	if ( n == 0 )
	{
		EnableNonBlockSock(pzftp, nFD, false);
		return true;
	}

	FD_ZERO(&rset);
	FD_SET((unsigned int)nFD, &rset);

	wset = rset;

	if ( (n = select(nFD + 1, &rset, &wset, NULL, &timeout)) == 0 )
	{
		EnableNonBlockSock(pzftp, nFD, false);
		sprintf(pzftp->m_szError, "In Connect() connect timeout %s:%d, %s", pzftp->m_szServer, nPort, GetSysErr(STDSOCKERR));
		return false;
	}

	if ( FD_ISSET(nFD, &rset) || FD_ISSET(nFD, &wset) )
	{
		int nErr = 0;
		socklen_t  nLen = sizeof(nErr);
		   
		if ( getsockopt(nFD, SOL_SOCKET, SO_ERROR, (char*)&nErr, &nLen) < 0 )
		{
			EnableNonBlockSock(pzftp, nFD, false);
			return false;
		}

		EnableNonBlockSock(pzftp, nFD, false);
		return true;
	}

	EnableNonBlockSock(pzftp, nFD, false);
	return false;
}

bool LOGIN(ZFTP *pzftp)
{
	if ( !StartUp(pzftp->m_szError) )
		return false;

	if ( pzftp->m_szServer[0] == 0 )
	{
		sprintf(pzftp->m_szError, "In LOGIN(), m_szServer is empty.");
		return false;
	}

	if ( pzftp->m_nPort <= 0 )
	{
		sprintf(pzftp->m_szError, "In LOGIN(), Invalid port number [%d]", pzftp->m_nPort);
		return false;
	}

	if ( pzftp->m_szUserID[0] == 0 )
	{
		sprintf(pzftp->m_szError, "In LOGIN(), m_szUserID is empty.");
		return false;
	}

	if ( pzftp->m_szUserPW[0] == 0 )
	{
		sprintf(pzftp->m_szError, "In LOGIN(), m_szUserPW is empty.");
		return false;
	}
	
	if ( (pzftp->m_nFD = socket(AF_INET, SOCK_STREAM, 0)) == -1 )
	{
		sprintf(pzftp->m_szError, "In LOGIN(), %s", GetSysErr(STDSOCKERR));
		return false;
	}

	GetMyIP(pzftp);

	if ( pzftp->m_lstLocalIP.m_nCount <= 0 )
		return false;

	sprintf(pzftp->m_szLocalIP, "%s", (const char*) pzftp->m_lstLocalIP.m_pHead->m_pData);
	
	if ( !Connect(pzftp) )
		return false;

	if ( !Greeting(pzftp) )
		return false;

	if ( !USER(pzftp) )
		return false;

	if ( !PASS(pzftp) )
		return false;

	if ( !SYST(pzftp) )
		return false;

	return true;
}

bool HasDDIR(const char *p)
{
	if ( !p )
		return false;

	if ( strchr(p, '/') != NULL )
		return true;

	if ( strchr(p, '\\') != NULL )
		return true;

	return false;
}

bool RecvResp(ZFTP *pzftp, const char *pszRetCode)
{
	int n = 0;
	char *p = NULL;
	char szNext[5];
	char szEOF[5];
	
	p = pzftp->m_pszLineBuffer;

	sprintf(szNext, "%s-", pszRetCode);
	sprintf(szEOF, "%s ", pszRetCode);
	
	if ( (n=RecvLine(pzftp, p)) < 0 )
		return false;
	
	PrintF(pzftp, "<< %s\n", p);
	if ( strstr(p, pszRetCode) != p )
	{
		sprintf(pzftp->m_szError, "bad response");
		return false;
	}

	if ( strstr(p, szEOF) == p )
		return true;

	if ( strstr(p, szNext) == p )
	{
		while ( RecvLine(pzftp, p) )
		{
			PrintF(pzftp, "<< %s\n", p);
			if ( strstr(p, szEOF) == p )
				return true;
		}
		
		sprintf(pzftp->m_szError, "bad response");
		return false;
	}

	return true;
}

bool Greeting(ZFTP *pzftp)
{
	return RecvResp(pzftp, "220");
}

bool USER(ZFTP *pzftp)
{
	int n = sprintf(pzftp->m_pszSendBuffer, "USER %s\r\n", pzftp->m_szUserID);
	
	if ( !SockWrite(pzftp, pzftp->m_nFD, pzftp->m_pszSendBuffer, n) )
		return false;

	return RecvResp(pzftp, "331");
}

bool PASS(ZFTP *pzftp)
{
	int i = 0, n = 0;
	
	PrintF(pzftp, ">> PASS ");
	n = strlen(pzftp->m_szUserPW);
	for ( i = 0; i < n; i++ )
		PrintF(pzftp, "*");

	PrintF(pzftp, "\n");
	
	n = sprintf(pzftp->m_pszSendBuffer, "PASS %s\r\n", pzftp->m_szUserPW);

	if ( !SockWrite(pzftp, pzftp->m_nFD, pzftp->m_pszSendBuffer, n) )
		return false;
	
	return RecvResp(pzftp, "230");
}

bool TYPE(ZFTP *pzftp, const bool bBinary)
{
	int n = 0;
	
	if ( bBinary ) 
		n = sprintf(pzftp->m_pszSendBuffer, "TYPE I\r\n");
	else
		n = sprintf(pzftp->m_pszSendBuffer, "TYPE A\r\n");
	
	printf(">> TYPE I\n");
	
	if ( !SockWrite(pzftp, pzftp->m_nFD, pzftp->m_pszSendBuffer, n) )
		return false;

	return RecvResp(pzftp, "200");
}

bool SYST(ZFTP *pzftp)
{
	int n = sprintf(pzftp->m_pszSendBuffer, "SYST\r\n");
	
	PrintF(pzftp, ">> SYST\n");

	if ( !SockWrite(pzftp, pzftp->m_nFD, pzftp->m_pszSendBuffer, n) )
		return false;

	if ( (n=RecvLine(pzftp, pzftp->m_pszLineBuffer)) < 0 )
		return false;

	if ( strstr(pzftp->m_pszLineBuffer, "215 ") != pzftp->m_pszLineBuffer )
	{
		sprintf(pzftp->m_szError, "In SYST(), invalid reponse [%s]", pzftp->m_pszLineBuffer);
		return false;
	}
	
	PrintF(pzftp, "<< %s\n", pzftp->m_pszLineBuffer);

	if ( Search(pzftp->m_pszLineBuffer, n, "UNIX", 4, true) >= 0 )
		pzftp->m_sysType = UNIX;
	else
		pzftp->m_sysType = WINDOWS;

	return true;
}

bool CWD(ZFTP *pzftp, const char *pszPath)
{
printf("In CWD[%s]\n", pszPath);
	if ( !Expect(pzftp, "250 ", 4, "CWD %s\r\n", pszPath) )
		return false;

	return true;
}

bool SmartCWD(ZFTP *pzftp, const char *pszPath)
{
	int nPos = 0;
	char DDIR = 0;
	const char *p = 0;
	char szPath[MAX_PATH];
	
	memset(szPath, 0x00, sizeof(szPath));

	DDIR = (char)((pzftp->m_sysType == UNIX) ? '/' : '\\');
	p = strrchr(pszPath, DDIR);

	nPos = p-pszPath;
	
	if ( nPos > 0 )
	{
		strncpy(szPath, pszPath, nPos);
		if ( !CWD(pzftp, szPath) )
			return false;
	}

	return true;
}

bool QUIT(ZFTP *pzftp)
{
	if ( !Expect(pzftp, "221 ", 4, "QUIT\r\n") )
		return false;

	return true;
}

bool CreatePort(ZFTP *pzftp, int *rpnSockFD, unsigned int *rpnPort)
{
	struct sockaddr_in sa;
	register int nRetryCnt = 0;

	*rpnSockFD = 0;
	*rpnPort = 2000 + (((rand()+10) % 5000) / 2);

	//////////////////////////////////////////////////////////////////////////
	// ȿ Ʈ ãƼ, Ʈ ȣ ٲپ õ.
	//////////////////////////////////////////////////////////////////////////
	for ( nRetryCnt = 0; nRetryCnt < MAX_RETRY_CREATE_PORT_SOCKET; nRetryCnt++ )
	{

		if ( ( *rpnSockFD = socket ( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
		{
			sprintf(pzftp->m_szError, "In CreatePort(), Can't create server socket, %s", GetSysErr(STDSOCKERR));
			return false;
		}

		SetReuse(*rpnSockFD, true);
		
		*rpnPort += 2;
		memset( &sa, 0, sizeof ( sa ) );
		sa.sin_family = AF_INET;
		sa.sin_addr.s_addr = htonl(INADDR_ANY);
		sa.sin_port = htons((unsigned short) *rpnPort);

		if ( bind(*rpnSockFD, (const struct sockaddr*)&sa, sizeof(sa)) != SOCKET_ERROR )
			break;

		closesocket(*rpnSockFD);
	}

	if ( listen (*rpnSockFD, 8) == SOCKET_ERROR )
	{
		sprintf(pzftp->m_szError, "Can't listen server socket, %s", GetSysErr(STDSOCKERR));
		return false;
	}

	return true;
}

void SetReuse(int nFD, bool bSet)
{
	int n;
	(bSet == true) ? (n = 1) : (n = 0);
	setsockopt(nFD, SOL_SOCKET, SO_REUSEADDR, (char *) &n, sizeof(n));
}

void SetTcpNoDelay(int nFD, bool bSet)
{
	int n;
	(bSet == true) ? (n = 1) : (n = 0);
	setsockopt(nFD, IPPROTO_TCP, TCP_NODELAY, (char *) &n, sizeof(int));
}

bool Pasv(ZFTP *pzftp, char *rpszIP, unsigned int *rpnPort)
{
	int n = 0, n1 = 0, n2 = 0;
	char *p = NULL;
	char *pTok = NULL;
	char szLine[255];

	PrintF(pzftp, ">> PASV\n");
	if ( !SockWrite(pzftp, pzftp->m_nFD, "PASV\r\n",6) )
		return false;

	if ( (n=RecvLine(pzftp, szLine)) <= 0 )
		return false;

	PrintF(pzftp, "<< %s\n", szLine);
	if ( strstr(szLine, "227") != szLine )
	{
		printf("bad response, %s\n", szLine);
		return false;
	}

	p = strchr(szLine, '(');
	if ( !p )
	{
		printf("bad response, %s\n", szLine);
		return false;
	}

	p++;
	if ( (pTok = strtok(p, ",")) == NULL )
		return false;
	else
		n = sprintf(rpszIP, "%s.", pTok);

	if ( (pTok=strtok(NULL, ",")) == NULL )
		return false;
	else 
		n+= sprintf(&rpszIP[n], "%s.", pTok);

	if ( (pTok=strtok(NULL, ",")) == NULL )
		return false;
	else
		n+= sprintf(&rpszIP[n], "%s.", pTok);

	if ( (pTok = strtok(NULL, ",")) == NULL )
		return false;
	else
		n+= sprintf(&rpszIP[n], "%s", pTok);

	if ( (pTok=strtok(NULL, ",")) == NULL )
		return false;

	n1 = atoi(pTok);	

	if ( (pTok=strtok(NULL, ",")) == NULL )
		return false;

	n2 = atoi(pTok);
	*rpnPort =  n1*256+n2;
	return true;
}

bool Port(ZFTP *pzftp, const unsigned int nPort)
{
	int i1, i2, i3, i4, n = 0;
	char szCmd[128];

	if ( sscanf(pzftp->m_szLocalIP, "%d.%d.%d.%d", &i1, &i2, &i3, &i4) != 4 )
	{
		sprintf(pzftp->m_szError, "In Port(), invalid ip format %s", pzftp->m_szLocalIP);
		return false;
	}
		
	n = sprintf(szCmd, "PORT %d,%d,%d,%d,", i1, i2, i3, i4);
	n += sprintf(&szCmd[n], "%d,%d\r\n", nPort/256, nPort%256);

	if ( !Expect(pzftp, "200 ", 4, "%s", szCmd) )
	{
		return false;
	}
	
	return true;
}

bool RETR(ZFTP *pzftp, const char *pszPath)
{
	if ( !Expect(pzftp, "150 ", 4, "RETR %s\r\n", pszPath) )
		return false;

	return true;
}

bool STOR(ZFTP *pzftp, const char *pszPath)
{
	if ( !Expect(pzftp, "150 ", 4, "STOR %s\r\n", pszPath) )
		return false;

	return true;
}

bool Download(ZFTP *pzftp, const char *pszRemotePath, const char *pszLocalPath, int64 *rpnTotRecvSize, const bool bPrintHashMark)
{
	char szFileName[MAX_PATH];
	unsigned int nPort = 0;
	int i = 0, n = 0, nSvrSockFD = 0;
	int nPeerFD = 0;
	FILE *fpOut = NULL;
	struct sockaddr_in peerAddr;
	socklen_t nAddrLen = sizeof(peerAddr);
	int nRecv = 0, nPrintCnt=0;
	bool bFirst = true;
	bool bPasv = true;
	char *pszBuffer = (char*)malloc(sizeof(char)*(MAX_SOCK_BUFFER+10));
	char szIP[32];

	memset(szIP, 0, sizeof(szIP));
	memset(szFileName, 0, MAX_PATH);
	memset(pszBuffer, 0, MAX_SOCK_BUFFER+10);

	if ( !SmartCWD(pzftp, pszRemotePath) )
		return false;
	
	if ( HasDDIR(pszRemotePath) )
	{
		ExtractFileName(pszRemotePath, szFileName);
	}
	else
	{
		sprintf(szFileName, "%s", pszRemotePath);
	}

	if ( !TYPE(pzftp, true) )
		return false;

	memset(&peerAddr, 0, nAddrLen);

	if ( bPasv )
	{
		if ( !Pasv(pzftp, szIP, &nPort) )
			return false;
		
		peerAddr.sin_family= AF_INET;
		peerAddr.sin_port = htons((unsigned short)nPort);
		peerAddr.sin_addr.s_addr = inet_addr(szIP);
		nPeerFD = socket(AF_INET, SOCK_STREAM, 0);

		if ( (nPeerFD = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
		{
			sprintf(pzftp->m_szError, "fail to create pasv socket, %s\n", strerror(errno));
			return false;
		}

		if ( connect(nPeerFD, (struct sockaddr *)&peerAddr, sizeof (struct sockaddr_in)) != 0 )
		{
			sprintf(pzftp->m_szError, "fail to connect %s:%d, %s\n", szIP, nPort, strerror(errno));
			return false;
		}
	}	
	else
	{
		for ( i = 0; i < 10; i++ )
		{
			if ( nSvrSockFD > 0 )
			{
				closesocket(nSvrSockFD);
				nSvrSockFD=0;
			}

			if ( !CreatePort(pzftp, &nSvrSockFD, &nPort) )
				continue;

			if ( !Port(pzftp, nPort) )
				continue;

			break;
		}

		if ( i == 10 )
			return false;
	}

	n = sprintf(pzftp->m_pszSendBuffer, "RETR %s\r\n", szFileName);
	if ( !SockWrite(pzftp, pzftp->m_nFD, pzftp->m_pszSendBuffer, n) )
		return false;

	if ( RecvLine(pzftp, pszBuffer) <= 0 )
		return false;

	if ( memicmp(pszBuffer, "150 ", 4) != 0 )
	{
		printf("bad [RETR] response, %s\n", pszBuffer);
		return false;
	}
	printf("<< %s\n", pszBuffer);

	if ( !bPasv )
		nPeerFD = accept(nSvrSockFD, (struct sockaddr *)&peerAddr, &nAddrLen);

	if ( nPeerFD == INVALID_SOCKET )
	{
		sprintf(pzftp->m_szError, "In Download(), fail to accept, %s", GetSysErr(STDSOCKERR));
		return false;
	}

	fpOut = fopen(pszLocalPath, "wb");
	if ( !fpOut )
	{
		sprintf(pzftp->m_szError, "In Download(), fail to open file %s, %s", pszLocalPath, strerror(errno));
		return false;
	}

	*rpnTotRecvSize = 0;
	
	while ( (nRecv=recv(nPeerFD, pszBuffer, MAX_SOCK_BUFFER, 0)) > 0 )
	{
		if ( bFirst )
		{
			printf("   ");
			bFirst = false;
		}

		if ( (int)fwrite(pszBuffer, 1, nRecv, fpOut) != nRecv )
		{
			sprintf(pzftp->m_szError, "In Download(), fail to write file, %s", strerror(errno));
			fclose(fpOut);
			closesocket(nPeerFD);
			free(pszBuffer);
			return false;
		}

		*rpnTotRecvSize += nRecv;
		if ( bPrintHashMark )
		{
			putchar('*');
			nPrintCnt++;
			if ( nPrintCnt > 49 )
			{
				printf("\n   ");
				nPrintCnt=0;
			}
		}
	}

	putchar('\n');

	fclose(fpOut);
	closesocket(nPeerFD);

	if ( !bPasv )
		closesocket(nSvrSockFD);

	if ( !RecvLine(pzftp, pszBuffer) )
		return false;

	if ( memicmp("226 ", pszBuffer, 4) == 0 )
	{
#ifdef WIN32
		printf(">> complete download [%s %I64d bytes]\n", pszLocalPath, *rpnTotRecvSize);
#else
		printf(">> complete download [%s %lld bytes]\n", pszLocalPath, *rpnTotRecvSize);
#endif
		free(pszBuffer);
		return true;
	}
	
	sprintf(pzftp->m_szError, "In Download(), invliad response [%s]", pszBuffer);
	free(pszBuffer);
	return false;
}

bool Upload(ZFTP *pzftp, const char *pszLocalPath, const char *pszRemotePath, const bool bPrintHashMark)
{
	char szFileName[MAX_PATH];
	unsigned int nPort = 0;
	int i = 0, n = 0, nSvrSockFD = 0;
	struct sockaddr_in peerAddr;
	socklen_t nAddrLen = sizeof(peerAddr);
	int nPeerFD = 0;
	FILE *fpIn = NULL;
	int nRead = 0, nPrintCnt = 0;
	bool bFirst = true;
	bool bPasv = true;
	int64 nTotSendSize = 0;
	char *pszBuffer = (char*)malloc(sizeof(char)*(MAX_SOCK_BUFFER+10));
    char szIP[32];

    memset(szIP, 0, sizeof(szIP));
	memset(szFileName, 0, MAX_PATH);
	memset(pszBuffer, 0, MAX_SOCK_BUFFER+10);

	if ( !TYPE(pzftp, true) )
	{
		free(pszBuffer);
		return false;
	}

	if ( !SmartCWD(pzftp, pszRemotePath) )
	{
		free(pszBuffer);
		return false;
	}

	if ( HasDDIR(pszRemotePath) )
	{
		ExtractFileName(pszRemotePath, szFileName);
	}
	else
	{
		sprintf(szFileName, "%s", pszRemotePath);
	}

    memset(&peerAddr, 0, nAddrLen);
    
    if ( bPasv )
    {
		if ( !Pasv(pzftp, szIP, &nPort) )
			return false;
		
		peerAddr.sin_family= AF_INET;
		peerAddr.sin_port = htons((unsigned short)nPort);
		peerAddr.sin_addr.s_addr = inet_addr(szIP);
		nPeerFD = socket(AF_INET, SOCK_STREAM, 0);

		if ( (nPeerFD = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
		{
			sprintf(pzftp->m_szError, "fail to create pasv socket, %s\n", strerror(errno));
			return false;
		}

		if ( connect(nPeerFD, (struct sockaddr *)&peerAddr, sizeof (struct sockaddr_in)) != 0 )
		{
			sprintf(pzftp->m_szError, "fail to connect %s:%d, %s\n", szIP, nPort, strerror(errno));
			return false;
		}
	}
	else
	{
    	for ( i = 0; i < 10; i++ )
    	{
    		if ( nSvrSockFD > 0 )
    		{
    			closesocket(nSvrSockFD);
    			nSvrSockFD=0;
    		}
    
    		if ( !CreatePort(pzftp, &nSvrSockFD, &nPort) )
    			continue;
    	
    		if ( !Port(pzftp, nPort) )
    			continue;
    
    		break;
    	}
    	
    	if ( i == 10 )
    	{
    		free(pszBuffer);
    		return false;
    	}
    }

	n = sprintf(pzftp->m_pszSendBuffer, "STOR %s\r\n", szFileName);
	if ( !SockWrite(pzftp, pzftp->m_nFD, pzftp->m_pszSendBuffer, n) )
		return false;

	if ( RecvLine(pzftp, pszBuffer) <= 0 )
		return false;

	if ( memicmp(pszBuffer, "150 ", 4) != 0 )
	{
		printf("bad [STOR] response, %s\n", pszBuffer);
		return false;
	}
	printf("<< %s\n", pszBuffer);

	if ( !bPasv ) 
	    nPeerFD = accept(nSvrSockFD, (struct sockaddr *)&peerAddr, &nAddrLen);

	if ( nPeerFD == INVALID_SOCKET )
	{
		sprintf(pzftp->m_szError, "In Upload(), fail to accept, %s", GetSysErr(STDSOCKERR));
		free(pszBuffer);
		return false;
	}

	// SetTcpNoDelay(nPeerFD, true);
	if ( (fpIn = fopen(pszLocalPath, "rb")) == NULL )
	{
		sprintf(pzftp->m_szError, "In Upload(), fail to open file, %s", strerror(errno));
		free(pszBuffer);
		return false;
	}

	while ( (nRead=fread(pszBuffer, 1, MAX_SOCK_BUFFER, fpIn)) > 0 )
	{		
		pszBuffer[nRead]=0;
		
		if ( bFirst )
		{
			printf("   ");
			bFirst = false;
		}

		if ( !SockWrite(pzftp, nPeerFD, pszBuffer, nRead) )
		{
			free(pszBuffer);
			fclose(fpIn);
			return false;
		}

		nTotSendSize += nRead;
		if ( bPrintHashMark )
		{
			putchar('*');
			fflush(stdout);
			nPrintCnt++;
			if ( nPrintCnt > 49 )
			{
				printf("\n   ");
				nPrintCnt=0;
			}
		}
	}
	
	putchar('\n');
	fclose(fpIn);
	closesocket(nPeerFD);
	
	if ( !bPasv )
	    closesocket(nSvrSockFD);
	
	if ( !RecvLine(pzftp, pszBuffer) )
	    return false;
	
	if ( memicmp("226 ", pszBuffer, 4) == 0 )
	{
#ifdef WIN32
		printf(">> upload success [%s %I64d bytes]\n", szFileName, nTotSendSize);
#else
		printf(">> upload success [%s %lld bytes]\n", szFileName, nTotSendSize);
#endif
		free(pszBuffer);
		return true;
	}

	sprintf(pzftp->m_szError, "In Upload(), invliad response [%s]", pszBuffer);
	free(pszBuffer);
	return false;
}

bool SockWrite(ZFTP *pzftp, int nFD, const char *p, const int nExpectWrite)
{
	int nCurWrite = 0, nTotWrite = 0;

	while ( Can(pzftp, nFD, WRITE) )
	{
		if ( nExpectWrite != nTotWrite && ((nCurWrite=send(nFD, p + nTotWrite, nExpectWrite - nTotWrite, 0)) > 0) )
		{
			nTotWrite += nCurWrite;

			if ( nTotWrite == nExpectWrite )
				break;
		}
		else
		{
			break;
		}
	}

	if ( nTotWrite < nExpectWrite )
	{
		sprintf(pzftp->m_szError, "In SockWrite(), fail to write %s\n", GetSysErr(STDSOCKERR));
		return false;
	}
	
	return true;
}

bool Expect(ZFTP *pzftp, const char *pExpect, int nExpect, const char *format, ...)
{
	int nExpectWrite = 0;
	int nCurWrite = 0, nTotWrite = 0;
	int nCurRecv = 0;
	va_list args;
	char szReq[255];

	va_start(args, format);
	nExpectWrite = vsprintf(pzftp->m_pszSendBuffer, format, args);
	va_end(args);

	sprintf(szReq, "%s", pzftp->m_pszSendBuffer);
	PrintF(pzftp, ">> %s", szReq);

	while ( nExpectWrite != nTotWrite && (nCurWrite=send(pzftp->m_nFD, pzftp->m_pszSendBuffer + nTotWrite, nExpectWrite - nTotWrite, 0)) > 0 )
		nTotWrite += nCurWrite;
	
	if ( nTotWrite < nExpectWrite )
	{
		sprintf(pzftp->m_szError, "In Expect, fail to write %s\n", GetSysErr(STDSOCKERR));
		return false;
	}

	nCurRecv = recv(pzftp->m_nFD, pzftp->m_pszSendBuffer, MAX_SOCK_BUFFER, 0);
	if ( nCurRecv < 0 )
	{
		sprintf(pzftp->m_szError, "In Expect, fail to recv, %s", GetSysErr(STDSOCKERR));
		return false;
	}

	pzftp->m_pszSendBuffer[nCurRecv] = 0;
	PrintF(pzftp, "<< %s", pzftp->m_pszSendBuffer);

	if ( memicmp(pExpect, pzftp->m_pszSendBuffer, nExpect) == 0 )
		return true;

	sprintf(pzftp->m_szError, "invalid response, request [%s], expect [%s], response [%s]\n", szReq, pExpect, pzftp->m_pszSendBuffer);
	return false;
}

void CloseSockFD(ZFTP *pzftp)
{
	if ( pzftp->m_nFD <= 0 )
		return;

#ifdef WIN32
	if ( closesocket(pzftp->m_nFD) < 0 )
#else
	if ( close(pzftp->m_nFD) < 0 )
#endif
	{
		sprintf(pzftp->m_szError, "In Close(), %s", GetSysErr(STDSOCKERR));
	}

	pzftp->m_nFD = 0;
}

void GetMyIP(ZFTP *pzftp)
{
	int i = 0;
	char buffer[256];
	struct hostent *pHostEnt;
	struct sockaddr_in localAddr;
	
	if ( gethostname(buffer, 256) != 0 )
	{
		sprintf(pzftp->m_szError, "In GetMyIP(), %s", GetSysErr(STDSOCKERR));
		return;
	}

	if((pHostEnt = gethostbyname(buffer)) != NULL)
	{
		for ( i = 0; ; i++ )
		{
			if ( pHostEnt->h_addr_list[i] == 0 )
				break;

			memset(&localAddr.sin_addr, 0, sizeof(localAddr.sin_addr));
			memcpy(&localAddr.sin_addr, pHostEnt->h_addr_list[i], pHostEnt->h_length);
			PushBack( &pzftp->m_lstLocalIP, strdup( inet_ntoa(localAddr.sin_addr) ) );
		}

		return;
	}
}

int Search(const char *pSrc, const int nSrcLen, const char *pFind, const int nFindLen, bool bCaseIgnore)
{
	int i = 0;
	if ( !pSrc || !pFind )
		return -1;
	
	if ( nSrcLen < nFindLen )
		return -1;

	if ( bCaseIgnore )
	{
		for ( i = 0; i < nSrcLen && i + nFindLen <= nSrcLen; i++ )
		{
			if ( toupper((int)pSrc[i]) == toupper((int)pFind[0]) && i + nFindLen <= nSrcLen )
			{
				if ( memicmp(&pSrc[i], pFind, nFindLen) == 0 )
					return i;
			}
		}

		return -1;
	}
	
	for ( i = 0; i < nSrcLen && i + nFindLen <= nSrcLen; i++ )
	{
		if ( pSrc[i] == pFind[0] && i + nFindLen <= nSrcLen )
		{
			if ( memcmp(&pSrc[i], pFind, nFindLen) == 0 )
				return i;
		}
	}	
	
	return -1;
}

bool Can(ZFTP *pzftp, unsigned int nFD, int nAct)
{
	int nRet = 0;

#ifdef WIN32
	struct timeval timeout;
	fd_set fdAct;
	fd_set fdErr;
#endif

	if ( nFD == -1 || !(nAct == READ || nAct == WRITE))
		return false;

#ifdef WIN32

	timeout.tv_sec = pzftp->m_nTimeout;
	timeout.tv_usec = 0;

	FD_ZERO(&fdAct);
	FD_ZERO(&fdErr);

	FD_SET(nFD, &fdAct);
	FD_SET(nFD, &fdErr);
	
    // On  success, select and pselect return the number of descriptors contained in the descriptor sets, which may be zero if the time-
    // out expires before anything interesting happens.  On error, -1 is returned, and errno is set appropriately; the sets and  timeout
    // become undefined, so do not rely on their contents after an error.
	nRet = ( nAct == READ ) ? select(nFD, &fdAct, NULL, NULL, &timeout) : select(nFD, NULL, &fdAct, &fdErr, &timeout);
	
	if (nRet > 0 && FD_ISSET(nFD, &fdAct) )
		return true;
#else

	struct pollfd po[1]={ {0} }; 
	const int nTimeout = pzftp->m_nTimeout * 1000;
	const int nPollEvent = (nAct == READ) ? (POLLIN | POLLPRI) : (POLLOUT | POLLWRBAND);

	po[0].fd  = nFD;
	po[0].events = nPollEvent;
	
	do
	{
		// [timeout] milliseconds. A negative value means infinite timeout
		//  On  error,  -1  is  returned,  and errno is set appropriately.
		nRet = poll (po, 1, nTimeout);
	} while ( nRet == -1 && errno == EINTR );

	if ( nRet == -1 )
	{
		switch ( errno )
		{
		case EBADF:		// An invalid file descriptor was given in one of the sets.
			return false;
		case ENOMEM:	// There was no space to allocate file descriptor tables.
			return false;
		case EFAULT:	// The array given as argument was not contained in the calling program's address space.
			return false;
		case EINVAL:	// The nfds value exceeds the RLIMIT_NOFILE value.
			return false;
		}
	}

	if ( po[0].fd > 0 && po[0].revents & nPollEvent )
		return true;

#endif
	
	if ( nRet < 0 )
	{
		sprintf(pzftp->m_szError, "In Can(), %s\n", GetSysErr(STDSOCKERR));
		return false;
	}

	return false;
}

#ifdef WIN32
bool InitCursor(ZFTP *pzftp, const char *pszDir, const char *pszFileExt)
{
	int n = 0;
	char szPath[MAX_PATH];

	if ( pzftp->m_hDir )
	{
		FindClose(pzftp->m_hDir);
		pzftp->m_hDir = NULL;
	}
	
	if ( pszDir == NULL )
	{
		sprintf(pzftp->m_szError, "In InitCursor(), pszDir is empty.");
		return false;
	}

	n = strlen(pszDir);
	
	if ( pszDir[n-1] == '\\' )
	{
		sprintf(szPath, "%s%s", pszDir, pszFileExt);
		sprintf(pzftp->m_szDir, "%s", pszDir);
	}
	else
	{
		sprintf(szPath, "%s\\%s", pszDir, pszFileExt);
		sprintf(pzftp->m_szDir, "%s\\", pszDir);
	}


	pzftp->m_hDir = FindFirstFile(szPath, &pzftp->m_findData);
	
	if ( pzftp->m_hDir == INVALID_HANDLE_VALUE )
	{
		sprintf(pzftp->m_szError, "In InitCursor(), fail to open such dir [%s], because %s", pszDir, strerror(errno));
		return false;
	}
	
	pzftp->m_bFirst = true;
	return true;
}

bool GetNextPath(ZFTP *pzftp, char *rpath)
{
	int n = 0;
	if ( pzftp->m_bFirst )
	{
		pzftp->m_bFirst = false;

		if ( pzftp->m_hDir == INVALID_HANDLE_VALUE )
			return false;

		if ( ( pzftp->m_findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != FILE_ATTRIBUTE_DIRECTORY )
		{
			n = sprintf(rpath, "%s", pzftp->m_szDir);
			sprintf(&rpath[n], "%s", pzftp->m_findData.cFileName);
			return true;
		}
	}

	n = sprintf(rpath, "%s", pzftp->m_szDir);

	while ( FindNextFile(pzftp->m_hDir, &pzftp->m_findData) )
	{
		if ( ( pzftp->m_findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == FILE_ATTRIBUTE_DIRECTORY )
			continue;

		sprintf(&rpath[n], "%s", pzftp->m_findData.cFileName);
		return true;
	}

	return false;
}

#else
bool InitCursor(ZFTP *pzftp, const char *pszDir, const char *pszFileExt)
{
	if ( pzftp->m_pDir )
	{
		closedir(pzftp->m_pDir);
		pzftp->m_pDir = NULL;
	}

	if ( pszDir[0] == 0 )
	{
		sprintf(pzftp->m_szError, "In InitCursor(), pszDir is empty.");
		return false;
	}
	
	sprintf(pzftp->m_szDir, "%s", pszDir);
	EnsureDDIR(pzftp->m_szDir, DDIR);

	if ( (pzftp->m_pDir = opendir(pzftp->m_szDir)) == NULL )
	{
		sprintf(pzftp->m_szError, "In InitCursor(), fail to open such dir[%s], because %s", pzftp->m_szDir, strerror(errno));
		return false;
	}
	return true;
}

bool GetNextPath(ZFTP *pzftp, char *rpath)
{
	struct dirent entry;
	struct dirent *result = NULL;
	struct stat st;
	
	//char szPath[MAX_PATH];

	register int n = sprintf(rpath, "%s", pzftp->m_szDir);
	while ( ( readdir_r(pzftp->m_pDir, &entry, &result) == 0 && result != NULL ) )
	{
		sprintf(&rpath[n], "%s", entry.d_name);
		
		if ( stat(rpath, &st) != 0 )
			continue;
		
		if ( S_ISDIR(st.st_mode) || strncmp(entry.d_name, ".", 1) == 0 || strncmp(entry.d_name, "..", 2) == 0 )
			continue;

		//if ( !Match(m_strExt, entry.d_name) )
		//	continue;

		//rstrPath = m_szPath;
		return true;
	}

	closedir(pzftp->m_pDir);
	pzftp->m_pDir = NULL;
	return false;
}

#endif

void EnsureDDIR(char *rpszPath, const char chDDIR)
{
	int n = strlen(rpszPath);
	if ( n <= 0 )
		return;

	if ( rpszPath[n-1] != chDDIR )
		rpszPath[n++] = chDDIR;

	rpszPath[n] = 0;
}


bool StartUp(char *rpszError)
{
#ifdef WIN32
	static bool bOnce = false;
	WSADATA wsaData;
	WORD wVersionRequested;

	if ( bOnce )
		return true;
	
	wVersionRequested = MAKEWORD(2,2);
	if ( WSAStartup(wVersionRequested, &wsaData) != 0 )
	{
		sprintf(rpszError, "%s", GetSysErr( (int)WSAGetLastError()));
		return false;
	}

	if ( LOBYTE(wsaData.wVersion) != 2 || HIBYTE(wsaData.wVersion) != 2 )
	{
		WSACleanup();
		sprintf(rpszError, "%s", GetSysErr( (int)WSAGetLastError() ));
		return false;
	}

	bOnce = true;
#endif
	return true;
}

int RecvLine(ZFTP *pzftp, char *rpszLine)
{
	int n = 0, nRecv = 0;
	const char *p = 0;

	if ( pzftp->m_nBufIdx <= 0 || pzftp->m_nBufIdx == pzftp->m_nBufLeng )
	{
		nRecv = recv(pzftp->m_nFD, pzftp->m_pszRecvBuffer, MAX_SOCK_BUFFER, 0);
		if ( nRecv <= 0 )
		{
			sprintf(pzftp->m_szError, "fail to RecvLine, %s\n", GetSysErr(STDSOCKERR));
			return -1;
		}

		pzftp->m_nBufLeng = nRecv;
		pzftp->m_pszRecvBuffer[nRecv] = 0;
		pzftp->m_nBufIdx = 0;
	}
	
	p = strchr(&pzftp->m_pszRecvBuffer[pzftp->m_nBufIdx], '\n');

	if ( !p )
	{
		n = sprintf(rpszLine, "%s", p);
		pzftp->m_nBufIdx = pzftp->m_nBufLeng;
	}
	else
	{
		//rstrLine = string(&pzftp->m_pszRecvBuffer[pzftp->m_nBufIdx], p - &pzftp->m_pszRecvBuffer[pzftp->m_nBufIdx]);
		n = p - &pzftp->m_pszRecvBuffer[pzftp->m_nBufIdx];
		strncpy(rpszLine, &pzftp->m_pszRecvBuffer[pzftp->m_nBufIdx], n);
		rpszLine[n] = 0;
		pzftp->m_nBufIdx += (n + 1);

	}

	return n;
}

const char *GetSysErr(int nErrNo)
{
	static char szError[1024];
#ifdef WIN32
	int nRet = 0;
	LPVOID lpMsgBuf=NULL;
#endif

	memset(szError, 0x00, 1024);

#ifdef WIN32
	nRet = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
					NULL, (DWORD)nErrNo, MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
					(LPTSTR)&lpMsgBuf, 0, NULL);

	if ( nRet == 0 )
		return szError;
		
	strncpy(szError, (LPCTSTR)lpMsgBuf, nRet);
	szError[nRet] = 0;

	if ( lpMsgBuf )
		LocalFree(lpMsgBuf);
	return szError;

#else 
	sprintf(szError, "%s", strerror(nErrNo));
#endif
	return szError;
}


void InitList(LLIST *pList)
{
	pList->m_pHead = NULL;
	pList->m_pTail = NULL;
	pList->m_nCount = 0;
}

void FreeList(LLIST *pList)
{
	PTRNODE *ptrNext = NULL;
	PTRNODE *ptrNode = pList->m_pHead;
	
	while ( ptrNode )
	{
		ptrNext = ptrNode->m_pNext;
		free(ptrNode->m_pData);
		free(ptrNode);
		ptrNode = ptrNext;
	}

	InitList(pList);
}

void PushBack(LLIST *pList, void *pData)
{
	PTRNODE *pNode = (PTRNODE *)malloc(sizeof(PTRNODE));
	pNode->m_pData = pData;
	pNode->m_pNext = NULL;
	pList->m_nCount++;

	if ( pList->m_pHead == NULL )
	{
		pList->m_pHead = pNode;
		pList->m_pTail = pNode;
		return;
	}
	
	pList->m_pTail->m_pNext = pNode;
	pList->m_pTail = pNode;
	pList->m_pTail->m_pNext = NULL;
}

void PrintF(ZFTP *pzftp, const char *format, ...)
{
	va_list args;

	if ( pzftp->m_bSilent )
		return;
		
	va_start(args, format);
	vprintf(format, args);
	va_end(args);
}

void WriteLog(ZFTP *pzftp, const char *format, ...)
{
	va_list args;
	va_start(args, format);
	vprintf(format, args);
	va_end(args);
}

