/**
	@file		ZFTP.h
	@author		aswip78@gmail.com
	@date		2008 / 02 / 06
	@version	v1.00.01
	
	@note

		- [2008.02.06]		Alpha test version 1.0 release.
*/

#ifndef ZFTP_H
#define ZFTP_H

#ifdef WIN32
#  pragma warning (disable:4786)
#  pragma comment (lib, "ws2_32.lib")
#endif

#ifdef WIN32
#  include <winsock2.h>
#  include <windows.h>
#  include <io.h>
#  include <conio.h>
#elif FREEBSD
#  include <stdio.h>
#  include <stdlib.h>
#  include <stdarg.h>
#  include <string.h>
#  include <ctype.h>
#  include <errno.h>
#  include <unistd.h>
#  include <sys/types.h>
#  include <sys/socket.h>
#  include <sys/poll.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>
#  include <netdb.h>
#  include <sys/event.h>
#  include <sys/time.h>
#  include <netinet/tcp.h>
#  include <unistd.h>
#elif AIX
#  include <unistd.h>
#  include <memory.h>
#  include <errno.h>
#  include <string.h>
#  include <stdio.h>
#  include <sys/signal.h>
#  include <fcntl.h>
#  include <arpa/inet.h>
#  include <sys/socket.h>
#  include <sys/poll.h>
#  include <sys/types.h>
#  include <netinet/tcp.h>
#  include <netinet/in.h>
#  include <netdb.h>
#  include <sys/time.h>
#  include <strings.h>
#else
#  include <sys/types.h>
#  include <dirent.h>
#  include <sys/socket.h>
#  include <netinet/tcp.h>
#  include <unistd.h>
#  include <netdb.h>
#  include <netinet/in.h>
#  include <netinet/tcp.h>
#  include <arpa/inet.h>
#  define __USE_XOPEN
#  include <sys/poll.h>
#  include <termios.h>
#  include <term.h>
#  include <ctype.h>
#  include <string.h>
#  include <stdlib.h>
#  include <errno.h>
#  include <memory.h>
#  include <pthread.h>
#endif

#ifdef WIN32
#  define STDSOCKERR		WSAGetLastError()
#else
#  define STDSOCKERR		errno
#endif

typedef int bool;

#ifndef true
#  define true	1
#endif

#ifndef false
#  define false	0
#endif

#ifndef __socklen_t_defined
#  ifndef WIN32 
typedef int socklen_t;
#  endif
#endif

#ifdef WIN32
#  define vsnprintf		_vsnprintf
#  define snprintf		_snprintf
   typedef int			socklen_t;
#else
#  define memicmp		strncasecmp
#endif

#ifndef MAX_PATH
#  define MAX_PATH		255
#endif

#define MAX_SOCK_BUFFER					4096
#define MAX_RETRY_CREATE_PORT_SOCKET	10

#if defined(WIN32) || defined(_MSC_VER) || defined(__BORLANDC__)
  // format character : %I64d	
  // possible values  : 0 ~ 18,446,744,073,709,551,615
  typedef __int64	int64;
  typedef unsigned __int64 uint64;
#else
  // format character : %lld
  // possible values  : 0 ~ 18,446,744,073,709,551,615 
  typedef long long int int64;
  typedef unsigned long long int uint64;
#endif

#ifndef INVALID_SOCKET
#  define INVALID_SOCKET  (unsigned int)(~0)
#endif

#ifndef SOCKET_ERROR
#  define SOCKET_ERROR            (-1)
#endif

typedef enum SOCKACT
{
	READ = 0,
	WRITE = 1,
} SOCKACT;

typedef enum SYSTEM_TYPE
{
	UNIX = 1,
	WINDOWS = 2
} SYSTEM_TYPE;

typedef enum COMMAND
{
	PUT = 1,
	GET = 2
} COMMAND;

struct PTRNODE;

typedef struct PTRNODE
{
	struct PTRNODE *m_pNext;
	void *m_pData;
}PTRNODE;

typedef struct LLIST
{
	PTRNODE *m_pHead;
	PTRNODE *m_pTail;
	int m_nCount;
} LLIST;

typedef struct PATH
{
	char m_szLocalPath[MAX_PATH];
	char m_szRemotePath[MAX_PATH];
} PATH;

typedef struct ZFTP
{
	int m_nBufIdx;
	int m_nBufLeng;
	bool m_bSilent;
	char m_szError[255];
	char *m_pszSendBuffer;
	char *m_pszRecvBuffer;
	char *m_pszLineBuffer;
	char m_szServer[128];
	char m_szLocalIP[32];
	unsigned short m_nPort;
	unsigned int m_nFD;
	int m_nTimeout;
	char m_szUserID[32];
	char m_szUserPW[32];
	struct sockaddr_in m_SockAddr;
	COMMAND m_cmd;
	SYSTEM_TYPE m_sysType;

	char m_szDir[MAX_PATH];
#ifdef WIN32
	bool m_bFirst;
	HANDLE m_hDir;
	WIN32_FIND_DATA m_findData;
	CRITICAL_SECTION m_Lock;
#else
	DIR *m_pDir;
	pthread_mutex_t m_Lock;
#endif
	LLIST m_lstLocalIP;
	LLIST m_lstPath;
} ZFTP;

#ifdef WIN32
	static const char DDIR = '\\';
#else
	static const char DDIR = '/';
#endif

/*
static const char *bstr = 
        "ABCDEFGHIJKLMNOPQ" 
        "RSTUVWXYZabcdefgh" 
        "ijklmnopqrstuvwxy" 
        "z0123456789+/";    

static const char rstr[] = {
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  62,   0,   0,   0,  63, 
         52,  53,  54,  55,  56,  57,  58,  59,  60,  61,   0,   0,   0,   0,   0,   0, 
          0,   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14, 
         15,  16,  17,  18,  19,  20,  21,  22,  23,  24,  25,   0,   0,   0,   0,   0, 
          0,  26,  27,  28,  29,  30,  31,  32,  33,  34,  35,  36,  37,  38,  39,  40, 
         41,  42,  43,  44,  45,  46,  47,  48,  49,  50,  51,   0,   0,   0,   0,   0};
*/

void InitList(LLIST *pList);
void FreeList(LLIST *pList);
void PushBack(LLIST *pList, void *pData);

void Init(ZFTP *pzftp);
void UnInit(ZFTP *pzftp);
void SetTimeout(ZFTP *pzftp, int nTimeout);
bool StartUp(char *rpszError);
const char *GetVerStr();
void SetServer(ZFTP *pzftp, const char *pszServer, const int nPort);
void SetAccount(ZFTP *pzftp, const char *pszUserID, const char *pszUserPW);
int PrintHelp(bool bDirectExit);
int PrintVer(bool bDirectExit);

bool IsDir(const char *pszPath);
void ExtractFileName(const char *pszPath, char *rpszFileName);
const char *GetLastErr(ZFTP *pzftp);
int Process(ZFTP *pzftp);
int EatPassword(char *rpszPass);
bool InitCursor(ZFTP *pzftp, const char *pszDir, const char *pszFileExt);
bool GetNextPath(ZFTP *pzftp, char *rpath);
void EnsureDDIR(char *rpszPath, const char chDDIR);
const char *GetSysErr(int nErrNo);
bool LOGIN(ZFTP *pzftp);
void GetMyIP(ZFTP *pzftp);
bool Connect(ZFTP *pzftp);
bool EnableBlockSock(ZFTP *pzftp, int nFD, bool bEnable);
void CloseSockFD(ZFTP *pzftp);
bool ConnectAddr(ZFTP *pzftp, int nFD, const char *pszServer, int nPort, struct sockaddr_in *psa);
bool Greeting(ZFTP *pzftp);
int RecvLine(ZFTP *pzftp, char *rpszLine);
bool RecvResp(ZFTP *pzftp, const char *pszRetCode);
bool HasDDIR(const char *p);
bool SockWrite(ZFTP *pzftp, int nFD, const char *p, const int nExpectWrite);
bool Can(ZFTP *pzftp, unsigned int nFD, int nAct);
bool Expect(ZFTP *pzftp, const char *pExpect, int nExpect, const char *format, ...);
bool USER(ZFTP *pzftp);
bool PASS(ZFTP *pzftp);
bool TYPE(ZFTP *pzftp, const bool bBinary);
bool SYST(ZFTP *pzftp);
int Search(const char *pSrc, const int nSrcLen, const char *pFind, const int nFindLen, bool bCaseIgnore);
bool CWD(ZFTP *pzftp, const char *pszPath);
bool SmartCWD(ZFTP *pzftp, const char *pszPath);
bool QUIT(ZFTP *pzftp);
void SetReuse(int nFD, bool bSet);
void SetTcpNoDelay(int nFD, bool bSet);
bool CreatePort(ZFTP *pzftp, int *rpnSockFD, unsigned int *rpnPort);
bool Port(ZFTP *pzftp, const unsigned int nPort);
bool RETR(ZFTP *pzftp, const char *pszPath);
bool STOR(ZFTP *pzftp, const char *pszPath);
bool Download(ZFTP *pzftp, const char *pszRemotePath, const char *pszLocalPath, int64 *rpnTotRecvSize, const bool bPrintHashMark);
bool Upload(ZFTP *pzftp, const char *pszLocalPath, const char *pszRemotePath, const bool bPrintHashMark);
void PrintF(ZFTP *pzftp, const char *format, ...);
void WriteLog(ZFTP *pzftp, const char *format, ...);
bool ExistFile(ZFTP *pzftp, const char *pszPath);
void Lock(ZFTP *pzftp);
void UnLock(ZFTP *pzftp);
#endif
